@@ -294,7 +294,7 @@ interpret_builtin (FcFormatContext *c,
 	if (new_str)
 	{
 	    FcStrBufString (buf, new_str);
-	    free (new_str);
+	    FcStrFree (new_str);
 	    return FcTrue;
 	}
 	else
@@ -358,7 +358,7 @@ skip_percent (FcFormatContext *c)
 
     /* skip an optional width specifier */
     if (strtol ((const char *) c->format, (char **) &c->format, 10))
-	/* don't care */;
+        {/* don't care */}
 
     if (!expect_char (c, '{'))
 	return FcFalse;
@@ -544,7 +544,7 @@ interpret_count (FcFormatContext *c,
 		 FcStrBuf        *buf)
 {
     int count;
-    FcPatternElt *e;
+    FcPatternIter iter;
     FcChar8 buf_static[64];
 
     if (!expect_char (c, '#'))
@@ -554,16 +554,9 @@ interpret_count (FcFormatContext *c,
 	return FcFalse;
 
     count = 0;
-    e = FcPatternObjectFindElt (pat,
-				FcObjectFromName ((const char *) c->word));
-    if (e)
+    if (FcPatternFindIter (pat, &iter, (const char *) c->word))
     {
-	FcValueListPtr l;
-	count++;
-	for (l = FcPatternEltValues(e);
-	     l->next;
-	     l = l->next)
-	    count++;
+	count = FcPatternIterValueCount (pat, &iter);
     }
 
     snprintf ((char *) buf_static, sizeof (buf_static), "%d", count);
@@ -695,7 +688,7 @@ interpret_simple (FcFormatContext *c,
 		  FcPattern       *pat,
 		  FcStrBuf        *buf)
 {
-    FcPatternElt *e;
+    FcPatternIter iter;
     FcBool        add_colon = FcFalse;
     FcBool        add_elt_name = FcFalse;
     int           idx;
@@ -743,9 +736,7 @@ interpret_simple (FcFormatContext *c,
 	c->word = orig;
     }
 
-    e = FcPatternObjectFindElt (pat,
-				FcObjectFromName ((const char *) c->word));
-    if (e || else_string)
+    if (FcPatternFindIter (pat, &iter, (const char *) c->word) || else_string)
     {
 	FcValueListPtr l = NULL;
 
@@ -757,8 +748,7 @@ interpret_simple (FcFormatContext *c,
 	    FcStrBufChar (buf, '=');
 	}
 
-	if (e)
-	    l = FcPatternEltValues(e);
+	l = FcPatternIterGetValues (pat, &iter);
 
 	if (idx != -1)
 	{
@@ -769,14 +759,14 @@ interpret_simple (FcFormatContext *c,
 	    }
 	    if (l && idx == 0)
 	    {
-		if (!FcNameUnparseValue (buf, &l->value, '\0'))
+		if (!FcNameUnparseValue (buf, &l->value, NULL))
 		    return FcFalse;
 	    }
 	    else goto notfound;
         }
 	else if (l)
 	{
-	    FcNameUnparseValueList (buf, l, '\0');
+	    FcNameUnparseValueList (buf, l, NULL);
 	}
 	else
 	{
@@ -790,7 +780,7 @@ interpret_simple (FcFormatContext *c,
 }
 
 static FcBool
-cescape (FcFormatContext *c,
+cescape (FcFormatContext *c FC_UNUSED,
 	 const FcChar8   *str,
 	 FcStrBuf        *buf)
 {
@@ -811,7 +801,7 @@ cescape (FcFormatContext *c,
 }
 
 static FcBool
-shescape (FcFormatContext *c,
+shescape (FcFormatContext *c FC_UNUSED,
 	  const FcChar8   *str,
 	  FcStrBuf        *buf)
 {
@@ -829,7 +819,7 @@ shescape (FcFormatContext *c,
 }
 
 static FcBool
-xmlescape (FcFormatContext *c,
+xmlescape (FcFormatContext *c FC_UNUSED,
 	   const FcChar8   *str,
 	   FcStrBuf        *buf)
 {
@@ -1014,7 +1004,7 @@ interpret_convert (FcFormatContext *c,
 	if (new_str)
 	{
 	    FcStrBufString (buf, new_str);
-	    free (new_str);
+	    FcStrFree (new_str);
 	    return FcTrue;
 	}
 	else
@@ -1193,12 +1183,19 @@ FcPatternFormat (FcPattern *pat,
 {
     FcStrBuf        buf;
     FcChar8         buf_static[8192 - 1024];
+    FcPattern      *alloced = NULL;
     FcBool          ret;
 
+    if (!pat)
+	alloced = pat = FcPatternCreate ();
+
     FcStrBufInit (&buf, buf_static, sizeof (buf_static));
 
     ret = FcPatternFormatToBuf (pat, format, &buf);
 
+    if (alloced)
+      FcPatternDestroy (alloced);
+
     if (ret)
 	return FcStrBufDone (&buf);
     else
